% Takes as input a real polynomial P and returns its roots,
% that is, the generalized eigenvalues of the associated
% companion pencil. P should be normalized w.r.t. 2-norm.
%
% This is an implementation of fast QZ for companion pencils
% with double shift ("A Real QZ Algorithm for Structured
% Companion Pencils", by P. Boito, Y. Eidelman and L. Gemignani).
%
% The pencil is A-lambda*B, where
% A=V-z*w' and B=U-p*q'.
% It is the companion pencil associated with the polynomial P.
%
% See lines 26-27 for choice of convergence criterion when deflating.
%
% **********************************
% Last modification August 13, 2016.
%
% paola.boito@unilim.fr
% 
% **********************************

function myeig=unitary_QZ_mixed_shift_function(P)

P=P/norm(P); % comment if unnecessary
max_local_iter=13; % max number of iteration per eigenvalue
% (will perform exceptional shift when max_local_iter is reached)

% Choose between the two convergence criteria:
conv='d'; % comparison with diagonal entries (as in the paper)  
% conv='f'; % comparison with fixed tolerance threshold*eps, see below

% initialize global and local iteration counters
numiter_global=0;  % n of iterations overall
numiter_local=0; % n of iteration for each eigenvalue
threshold=1; % the tolerance in the stopping criterion is threshold*eps

% determine degree of polynomial P
N=length(P)-1;
% find generators for matrix pencil
[sigmaa,gv,hv,bv,z,w,db,gu,hu,bu,p,q]=companion_generators(P);  % or
% [sigmaa,gv,hv,bv,z,w,db,gu,hu,bu,p,q]=companion_generators_col(P);

% build pencil for later checks
% A=[-P(2:N+1);[eye(N-1),zeros(N-1,1)]];
% B=eye(N);
% B(1,1)=P(1);

% initialize variable for computed eigenvalues 
myeig=zeros(N,1);

% EXTERNAL LOOP (deflation): determine kin and kfin
% (the QZ sweep goes from index kin to index kfin).
% Only deflation on A for now.

kfin=N;  
conv_flag=0;
while (conv_flag==0)
    defl_flag=0;
    while defl_flag==0
        
    % Perform double deflation if applicable.
    if conv=='f' 
        tolerance=eps*threshold;
    elseif conv=='d'
        d1=gv(kfin-2,:)*hv(:,kfin-2)-z(kfin-2)*w(kfin-2)';
        d2=gv(kfin-1,:)*hv(:,kfin-1)-z(kfin-1)*w(kfin-1)';
        d3=gv(kfin,:)*hv(:,kfin)-z(kfin)*w(kfin)';
        tolerance=eps*(abs(d1)+abs(d2));
    else
        fprintf('What is the convergence criterion?\n')
    end
    
    if (abs(sigmaa(kfin-2))<tolerance)
        
        % fprintf('Deflation of a 2x2 block\n')
        a33 = gv(kfin-1,:)*hv(:,kfin-1)-z(kfin-1)*w(kfin-1)';
        a34 = gv(kfin-1,:)*bv(:,:,kfin-1)*hv(:,kfin)-z(kfin-1)*w(kfin)';
        a43 = sigmaa(kfin-1);
        a44 = (gv(kfin,:)*hv(:,kfin)-z(kfin)*w(kfin)');
        b33 = db(kfin-1);
        b44 = db(kfin);
        b34 = gu(kfin-1)*hu(kfin)-p(kfin-1)*q(kfin)';
        myeig(kfin-1:kfin)=eig([a33 a34; a43 a44],[b33 b34; 0 b44]);
        kfin=kfin-2;
        if kfin<=2
            conv_flag=1;
            defl_flag=1;
        end
        numiter_local=0;
        
    % Perform single deflation if applicable.
    
    if conv=='d'
        tolerance=eps*(abs(d2)+abs(d3));
    end
    
    elseif (abs(sigmaa(kfin-1))<tolerance)
    
        % fprintf('Deflation of an 1x1 block\n')
        a44 = (gv(kfin,:)*hv(:,kfin)-z(kfin)*w(kfin)');
        b44 = db(kfin);
        myeig(kfin)=a44/b44;
        kfin=kfin-1;
        if kfin<=2
            conv_flag=1;
            defl_flag=1;
        end
        numiter_local=0;
    else
        defl_flag=1; % done enough deflations
    end
    end
    if conv_flag==0
    % Decide whether to apply single or double shift.
    % Compute 2x2 trailing matrix.
    a33 = (gv(kfin-1,:)*hv(:,kfin-1)-z(kfin-1)*w(kfin-1)');
    a34 = gv(kfin-1,:)*bv(:,:,kfin-1)*hv(:,kfin)-z(kfin-1)*w(kfin)';
    a43 = sigmaa(kfin-1);
    a44 = (gv(kfin,:)*hv(:,kfin)-z(kfin)*w(kfin)');
    b33 = db(kfin-1);
    b44 = db(kfin);
    b34 = gu(kfin-1)*hu(kfin)-p(kfin-1)*q(kfin)';
    C=[a33 a34; a43 a44]/[b33 b34; 0 b44];
    delta=(C(1,1)+C(2,2))^2-4*(C(1,1)*C(2,2)-C(1,2)*C(2,1));
    
    if delta>0
        % *************
        % single shift
        % *************
        
        % choose kin with single subdiag criterion
        
        kin=kfin-1;
        %while (kin>1)&&(abs(A(kin,kin-1))>tol)
        while (kin>1)&&(abs(sigmaa(kin-1))>eps*threshold)
        %while (kin>1)&&(abs(sigmaa(kin-1))>eps*min(abs(gv(kin,:)*hv(:,kin)-z(kin)*w(kin)')+abs(gv(kin,:)*hv(:,kin)-z(kin)*w(kin)'),1))
        %while (kin>1)&&(abs(sigmaa(kin-1))>eps*(abs(gv(kin,:)*hv(:,kin)-z(kin)*w(kin)')+abs(gv(kin,:)*hv(:,kin)-z(kin)*w(kin)')))
            kin=kin-1;
        end
        
        j=kfin-1;
        
        if numiter_local<=max_local_iter
            %xx=[B(kfin,kfin)*A(j,j)-A(kfin,kfin)*B(j,j);B(kfin,kfin)*A(j+1,j)];
            xx=[db(kfin)*(gv(j,:)*hv(:,j)-z(j)*w(j)')-(gv(kfin,:)*hv(:,kfin)-z(kfin)*w(kfin)')*db(j);db(kfin)*sigmaa(j)];
        else
            fprintf('exceptional shift\n')
            xx=rand(2,1);
        end
        G=planerot(xx);
        
        if j>1 % do the magic trick, otherwise we already know kin=1
            %yy=G*A(j:j+1,j-1);
            yy=G*[sigmaa(j-1);0];
            
            while (j>kin)&&(abs(yy(2))>eps*threshold)  
                j=j-1; % candidate to become kin
                % choose shift
                if numiter_local<=max_local_iter
                    %xx=[B(kfin,kfin)*A(j,j)-A(kfin,kfin)*B(j,j);B(kfin,kfin)*A(j+1,j)];
                    xx=[db(kfin)*(gv(j,:)*hv(:,j)-z(j)*w(j)')-(gv(kfin,:)*hv(:,kfin)-z(kfin)*w(kfin)')*db(j);db(kfin)*sigmaa(j)];
                    G=planerot(xx);
                end
                % check that column j-1 exists and compute yy
                if j>1
                    %yy=G*A(j:j+1,j-1);
                    yy=G*[sigmaa(j-1);0];
                end
                
            end
            % reinitialize counter after exceptional shift
            if numiter_local>max_local_iter
                numiter_local=0;
            end
            % update kin
            kin=j;
        else
            kin=1;
        end
        
        % perform single shift structured QZ from kin to kfin
        [sigmaa,gv,hv,bv,z,w,db,gu,hu,bu,p,q]=structured_QZ_ss(sigmaa,gv,hv,bv,z,w,db,gu,hu,bu,p,q,kin,kfin,G);
        
    else
        % *************
        % double shift
        % *************
        
        kin=kfin-1; % bulge chasing starts at index kin
        % (kin should be the largest possible index where we can start)
        kin_flag=0;
        while (kin_flag==0)&&(kin>1)
            kin=kin-1;
            % now see if this kin can work (magic trick)
            
            
            % method 3 for shift computation
            %
            %ascale=1/norm(A); % to be updated later if necessary
            %bscale=1/norm(B);
            ascale=1;
            bscale=1;
            a11 = (gv(kin,:)*hv(:,kin)-z(kin)*w(kin)');
            a21 = sigmaa(kin);
            a32 = sigmaa(kin+1);
            a12 = (gv(kin,:)*bv(:,:,kin)*hv(:,kin+1)-z(kin)*w(kin+1)');
            a22 = (gv(kin+1,:)*hv(:,kin+1)-z(kin+1)*w(kin+1)');
            a33 = (gv(kfin-1,:)*hv(:,kfin-1)-z(kfin-1)*w(kfin-1)');
            a34 = gv(kfin-1,:)*bv(:,:,kfin-1)*hv(:,kfin)-z(kfin-1)*w(kfin)';
            a43 = sigmaa(kfin-1);
            a44 = (gv(kfin,:)*hv(:,kfin)-z(kfin)*w(kfin)');
            b11 = db(kin);
            b22 = db(kin+1);
            b33 = db(kfin-1);
            b44 = db(kfin);
            b12 = gu(kin)*hu(kin+1)-p(kin)*q(kin+1)';
            b34 = gu(kfin-1)*hu(kfin)-p(kfin-1)*q(kfin)';
            
            ad11 = ascale*a33/(bscale*b33);
            ad21 = ascale*a43/(bscale*b33);
            ad12 = ascale*a34/(bscale*b44);
            ad22 = ascale*a44/(bscale*b44);
            u12 = b34/b44;
            ad11l = ascale*a11/(bscale*b11);
            ad21l = ascale*a21/(bscale*b11);
            ad12l = ascale*a12/(bscale*b22);
            ad22l = ascale*a22/(bscale*b22);
            ad32l = ascale*a32/(bscale*b22);
            u12l = b12/b22;
            a=zeros(3,1);
            a(1) = (ad11-ad11l)*(ad22-ad11l)-ad12*ad21+ad21*u12*ad11l+(ad12l-ad11l*u12l)*ad21l;
            a(2) = ((ad22l-ad11l)-ad21l*u12l-(ad11-ad11l)-(ad22-ad11l)+ad21*u12)*ad21l;
            a(3) = ad32l*ad21l;     
            
            initialQ=(hh3(a))';
            if kin>1
                start_vector=initialQ'*[sigmaa(kin-1);0;0]; % this will always have
                % the third entry equal to zero, by the way
                if norm(start_vector(2:3))<eps*threshold
                    kin_flag=1;
                end
            end
        end
        
        % implement exceptional shift
        if numiter_local>max_local_iter
            fprintf('Exceptional shift.\n');
            numiter_local=0;
            a=rand(3,1);
            initialQ=(hh3(a))';
            % apply magic trick to find kin
            kin=kfin-1;
            kin_flag=0;
            while (kin_flag==0)&&(kin>1)
                kin=kin-1;
                if kin>1
                    start_vector=initialQ'*[sigmaa(kin-1);0;0];
                    if norm(start_vector(2:3))<eps*threshold
                        kin_flag=1;
                    end
                end
            end
        end % end if-then for exceptional shift
        
        
        % now we know kin and kfin
        % let's perform bulge chasing from kin to kfin
        % (actually we perform it on the whole matrix,
        % except that Q and Z are set to the identity
        % when possible)
        
       [sigmaa1,gvnew,hvnew,bvnew,z1,w1,db1,gunew,hunew,bunew,p1,q1]=...
           structured_QZ_ds(sigmaa,gv,hv,bv,z,w,db,gu,hu,bu,p,q,kin,kfin,initialQ);
        
        % update generators
        gv=gvnew;
        hv=hvnew;
        bv=bvnew;
        sigmaa=sigmaa1;
        z=z1;
        w=w1;
        
        gu=gunew;
        hu=hunew;
        bu=bunew;
        db=db1;
        p=p1;
        q=q1;
        
    end % goes with choice of single or double shift
    
    numiter_global=numiter_global+1;
    numiter_local=numiter_local+1;
    % end iteration
    
    % compare to unstructured QZ with same Q and Z
    % (see check_normA and check_normB)
    % for jj=1:N-2
    %     A(jj:jj+2,:)=Q(:,:,jj)'*A(jj:jj+2,:);
    %     B(jj:jj+2,:)=Q(:,:,jj)'*B(jj:jj+2,:);
    %     A(:,jj:jj+2)=A(:,jj:jj+2)*Z(:,:,jj);
    %     B(:,jj:jj+2)=B(:,jj:jj+2)*Z(:,:,jj);
    % end
    % A(N-1:N,:)=Q_Nm1'*A(N-1:N,:);
    % B(N-1:N,:)=Q_Nm1'*B(N-1:N,:);
    % A(:,N-1:N)=A(:,N-1:N)*Z_Nm1;
    % B(:,N-1:N)=B(:,N-1:N)*Z_Nm1;
    
    
    % AAA=compute_matrix_A(gv,hv,bv,sigmaa,z,w);
    % BBB=compute_matrix_B(db,gu,hu,bu,p,q);
    
    %check_normA=norm(A-AAA);
    %check_normB=norm(B-BBB);
    
    % end of big loop
    
    end
end
if kfin==1
    myeig(1)=(gv(1,:)*hv(:,1)-z(1)*w(1)')/db(1);
elseif kfin==2
     a11 = (gv(1,:)*hv(:,1)-z(1)*w(1)');
     a21 = sigmaa(1);
     a12 = (gv(1,:)*bv(:,:,1)*hv(:,2)-z(1)*w(2)');
     a22 = (gv(2,:)*hv(:,2)-z(2)*w(2)');
     b11 = db(1);
     b22 = db(2);
     b12 = gu(1)*hu(2)-p(1)*q(2)';
    myeig(1:2)=eig([a11 a12; a21 a22],[b11 b12; 0 b22]);
else
    fprintf('Something is wrong.../n')
end